local _, iTF = ...
iTF.barTex = 'Interface\\Buttons\\WHITE8x8'
iTF.font = 'Fonts\\ARIALN.TTF'
iTF.fontSize = 11
local L = LibStub('AceLocale-3.0'):GetLocale('iTargetingFrames', true)
local specID = {
	id = 0,
}
local conditionals = {
	['onUpdate'] = {},
	['onAura'] = {},
	['onCast'] = {},
	['onHealth'] = {},
	['targetChanged'] = {},
	['focusUpdate'] = {},
	['threat'] = {},
}
local customConditionals = {}
iTF.indicatorUpdateFuncs = {
	['interruptRange'] = 'onUpdate',
	['currentTarget'] = 'targetChanged',
	['outOfCombat'] = 'onUpdate',
	['maxRange'] = 'onUpdate',
	['maxRangeDPS'] = 'onUpdate',
	['losingAggro'] = 'threat',
	['aggro'] = 'threat',
	['gainingAggro'] = 'threat',
	['focusTarget'] = 'focusUpdate',
}
local indicatorPrios = {}
iTF.backdrop = {
	bgFile = 'Interface\\Buttons\\WHITE8x8',
	edgeFile = 'Interface\\Buttons\\WHITE8x8',
	edgeSize = 1,
	insets = {
		left = -1,
		right = -1,
		top = -1,
		bottom = -1,
	}
}
local addon = CreateFrame('frame')
addon:RegisterEvent('ADDON_LOADED')
addon:SetScript('OnEvent', function(self, event, ...)
	self[event](self, ...)
end)
local function spairs(t, order)
    -- collect the keys
    local keys = {}
    for k in pairs(t) do keys[#keys+1] = k end

    -- if order function given, sort by it by passing the table and keys a, b,
    -- otherwise just sort the keys
    if order then
        table.sort(keys, function(a,b) return order(t, a, b) end)
    else
        table.sort(keys)
    end

    -- return the iterator function
    local i = 0
    return function()
        i = i + 1
        if keys[i] then
            return keys[i], t[keys[i]]
        end
    end
end
local function getFunction(str, unitID)
	local f, err = loadstring('return ' .. str .. ' ', unitID)
	if f then return f() else return f, err end
end
local function updateIndicator(unitID, cond, customCondIndicators, showCustom)
	local function getHighest(t)
		local w = -100
		local c
		for k,v in pairs(t) do
			if v.weight > w then
				w = v.weight
				c = k
			end
		end
		return w,c
	end
	local function border(color, oldUnitID, hideCurrentCond)
		local cond = cond
		local unitID = unitID
		if oldUnitID then
			unitID = oldUnitID
		end
		if iTF.frames[unitID].borderInUse == cond and not hideCurrentCond then
			return
		end
		if hideCurrentCond then
			iTF.frames[unitID].waitingFor.border[cond] = nil
			local w,c = getHighest(iTF.frames[unitID].waitingFor.border)
			if c then
				cond = c
				color = iTF.frames[unitID].waitingFor.border[cond].color
				iTF.frames[unitID].borderInUse = cond
			else
				iTF.frames[unitID].borderInUse = false
			end
		else
			iTF.frames[unitID].waitingFor.border[cond] = {
				['weight'] = indicatorPrios[cond],
				['color'] = color
			}
			local w,c = getHighest(iTF.frames[unitID].waitingFor.border)
			if c == iTF.frames[unitID].borderInUse then
				return
			end
			cond = c
			color = iTF.frames[unitID].waitingFor.border[cond].color
			iTF.frames[unitID].borderInUse = cond
		end
		if iTF.frames[unitID].borderInUse then
			iTF.frames[unitID]:SetBackdropBorderColor(unpack(color))
		else
			iTF.frames[unitID]:SetBackdropBorderColor(unpack(iTFConfig.layout.colors.backdrop.border))
		end
	end
	local function opacity(alpha, oldUnitID, hideCurrentCond)
		local cond = cond
		local unitID = unitID
		if oldUnitID then
			unitID = oldUnitID
		end
		if iTF.frames[unitID].alphaInUse == cond and not hideCurrentCond then
			return
		end
		if hideCurrentCond then
			iTF.frames[unitID].waitingFor.alpha[cond] = nil
			local w,c = getHighest(iTF.frames[unitID].waitingFor.alpha)
			if c then
				cond = c
				alpha = iTF.frames[unitID].waitingFor.alpha[cond].alpha
				iTF.frames[unitID].alphaInUse = cond
			else
				iTF.frames[unitID].alphaInUse = false
			end
		else
			iTF.frames[unitID].waitingFor.alpha[cond] = {
				['weight'] = indicatorPrios[cond],
				['alpha'] = alpha
			}
			local w,c = getHighest(iTF.frames[unitID].waitingFor.alpha)
			if c == iTF.frames[unitID].alphaInUse then
				return
			end
			cond = c
			alpha = iTF.frames[unitID].waitingFor.alpha[cond].alpha
			iTF.frames[unitID].alphaInUse = cond
		end
		if iTF.frames[unitID].alphaInUse then
			iTF.frames[unitID]:SetAlpha(alpha or 1)
		else
			iTF.frames[unitID]:SetAlpha(1)
		end
	end
	local function glows(k,color,oldUnitID, hideCurrentCond)
		local cond = cond
		local unitID = unitID
		if oldUnitID then
			unitID = oldUnitID
		end
		if iTF.frames[unitID].topFrame[k].inUse == cond and not hideCurrentCond then
			return
		end
		if hideCurrentCond then
			iTF.frames[unitID].waitingFor[k][cond] = nil
			local w,c = getHighest(iTF.frames[unitID].waitingFor[k])
			if c then
				cond = c
				color = iTF.frames[unitID].waitingFor[k][cond].color
				iTF.frames[unitID].topFrame[k].inUse = cond
			else
				iTF.frames[unitID].topFrame[k].inUse = false
			end
		else
			iTF.frames[unitID].waitingFor[k][cond] = {
				['weight'] = indicatorPrios[cond],
				['color'] = color
			}
			local w,c = getHighest(iTF.frames[unitID].waitingFor[k])
			if c == iTF.frames[unitID].topFrame[k].inUse then
				return
			end
			cond = c
			color = iTF.frames[unitID].waitingFor[k][cond].color
			iTF.frames[unitID].topFrame[k].inUse = cond
		end
		if iTF.frames[unitID].topFrame[k].inUse then
			iTF.frames[unitID].topFrame[k]:SetVertexColor(unpack(color))
			iTF.frames[unitID].topFrame[k]:Show()
		else
			iTF.frames[unitID].topFrame[k]:Hide()
		end
		
		
		--[[
		local unitID = unitID
		if oldUnitID then
			unitID = oldUnitID
		end
		local usedBy = iTF.frames[unitID].topFrame[k].inUse
		if color then
			if (not usedBy) or (usedBy and (cond ~= usedBy)) then
				if (not usedBy) or (usedBy and (indicatorPrios[cond] > indicatorPrios[usedBy])) then
					iTF.frames[unitID].topFrame[k].inUse = cond
					iTF.frames[unitID].topFrame[k]:SetVertexColor(unpack(color))
					iTF.frames[unitID].topFrame[k]:Show()
				end
			end
		elseif usedBy == cond then
			iTF.frames[unitID].topFrame[k].inUse = false
			iTF.frames[unitID].topFrame[k]:Hide()
		end
		--]]
	end
	local function statusbar(color, oldUnitID, hideCurrentCond)
		local cond = cond
		local unitID = unitID
		if oldUnitID then
			unitID = oldUnitID
		end
		if iTF.frames[unitID].healthBar.inUse == cond and not hideCurrentCond then
			return
		end
		if hideCurrentCond then
			iTF.frames[unitID].waitingFor.statusbar[cond] = nil
			local w,c = getHighest(iTF.frames[unitID].waitingFor.statusbar)
			if c then
				cond = c
				color = iTF.frames[unitID].waitingFor.statusbar[cond].color
				iTF.frames[unitID].healthBar.inUse = cond
			else
				iTF.frames[unitID].healthBar.inUse = false
			end
		else
			iTF.frames[unitID].waitingFor.statusbar[cond] = {
				['weight'] = indicatorPrios[cond],
				['color'] = color
			}
			local w,c = getHighest(iTF.frames[unitID].waitingFor.statusbar)
			if c == iTF.frames[unitID].healthBar.inUse then
				return
			end
			cond = c
			color = iTF.frames[unitID].waitingFor.statusbar[cond].color
			iTF.frames[unitID].healthBar.inUse = cond
		end
		if iTF.frames[unitID].healthBar.inUse then
			iTF.frames[unitID].healthBar:SetStatusBarColor(unpack(color))
		else
			if iTFConfig.layout.colors.classColor and iTF.frames[unitID].isPlayer then
				--local color = RAID_CLASS_COLORS[iTF.frames[unitID].isPlayer]
				--local color = iTF.frames[unitID].isPlayer.color
				iTF.frames[unitID].healthBar:SetStatusBarColor(iTF.frames[unitID].isPlayer.color.r, iTF.frames[unitID].isPlayer.color.g, iTF.frames[unitID].isPlayer.color.b, iTFConfig.layout.colors.statusbar.main.a)
			else
				iTF.frames[unitID].healthBar:SetStatusBarColor(unpack(iTFConfig.layout.colors.statusbar.main))
			end
		end
	end
	if cond == 'onUpdate' then
		--Hide all before showing them again
	end
	if cond == 'interruptRange' then
		if conditionals.onUpdate.interruptRange then
			if specID.interrupt then
				local show = false
				if conditionals.onUpdate.interruptRange.invert then
					if IsSpellInRange(specID.interrupt, unitID) == 0 then
						show = true
					end
				elseif IsSpellInRange(specID.interrupt, unitID) == 1 then
					show = true
				end
				for k in pairs(conditionals.onUpdate.interruptRange.indicators) do
					if k == 'border' then
						if show then
							border(conditionals.onUpdate.interruptRange.color)
						else
							border(nil, nil, true)
						end
					elseif k == 'alpha' then
						if show then
							opacity(conditionals.onUpdate.interruptRange.alpha)
						else
							opacity(nil, nil, true)
						end
					elseif k == 'statusbar' then
						if show then
							statusbar(conditionals.onUpdate.interruptRange.color)
						else
							statusbar(nil, nil, true)
						end
					elseif string.find(k, 'glow') then
						if show then
							glows(k,conditionals.onUpdate.interruptRange.color)
						else
							glows(k, nil, nil, true)
						end
					end
				end
			end
		end
	end
	if cond == 'maxRangeDPS' then
		if conditionals.onUpdate.maxRangeDPS then
			if specID.dps then
				local show = false
				if conditionals.onUpdate.maxRangeDPS.invert then
					if IsSpellInRange(specID.dps, unitID) == 0 then
						show = true
					end
				elseif IsSpellInRange(specID.dps, unitID) == 1 then
					show = true
				end
				for k in pairs(conditionals.onUpdate.maxRangeDPS.indicators) do
					if k == 'border' then
						if show then
							border(conditionals.onUpdate.maxRangeDPS.color)
						else
							border(nil, nil, true)
						end
					elseif k == 'alpha' then
						if show then
							opacity(conditionals.onUpdate.maxRangeDPS.alpha)
						else
							opacity(nil, nil, true)
						end
					elseif k == 'statusbar' then
						if show then
							statusbar(conditionals.onUpdate.maxRangeDPS.color)
						else
							statusbar(nil, nil, true)
						end
					elseif string.find(k, 'glow') then
						if show then
							glows(k,conditionals.onUpdate.maxRangeDPS.color)
						else
							glows(k, nil, nil, true)
						end
					end
				end
			end
		end
	end
	if cond == 'outOfCombat' then
		if conditionals.onUpdate.outOfCombat and unitID then
			local show = false
			if not iTF.encounterInProgress and not UnitAffectingCombat(unitID) then
				show = true
			end
			for k in pairs(conditionals.onUpdate.outOfCombat.indicators) do
				if k == 'border' then
					if show then
						border(conditionals.onUpdate.outOfCombat.color)
					else
						border(nil, nil, true)
					end
				elseif k == 'alpha' then
					if show then
						opacity(conditionals.onUpdate.outOfCombat.alpha)
					else
						opacity(nil, nil, true)
					end
				elseif k == 'statusbar' then
					if show then
						statusbar(conditionals.onUpdate.outOfCombat.color)
					else
						statusbar(nil, nil, true)
					end
				elseif string.find(k, 'glow') then
					if show then
						glows(k,conditionals.onUpdate.outOfCombat.color)
					else
						glows(k, nil, nil, true)
					end
				end
			end
		end
	end
	if cond == 'currentTarget' then
		if conditionals.targetChanged.currentTarget then
			local show = false
			if UnitExists('target') then
				local nameplate = C_NamePlate.GetNamePlateForUnit('target')
				if nameplate then
					unitID = nameplate.UnitFrame.unit
					show = true
				end
			end
			for k in pairs(conditionals.targetChanged.currentTarget.indicators) do
				if k == 'border' then
					if iTF.currentTarget then
						border(nil, iTF.currentTarget, true) --Hide old
					end
					if show then
						border(conditionals.targetChanged.currentTarget.color)
					elseif unitID then
						border(nil, nil, true)
					end
				elseif k == 'alpha' then
					if iTF.currentTarget then
						opacity(nil,iTF.currentTarget, true) --Hide old
					end
					if show then
						opacity(conditionals.targetChanged.currentTarget.alpha)
					elseif unitID then
						opacity(nil, nil, true)
					end
				elseif k == 'statusbar' then
					if iTF.currentTarget then
						statusbar(nil,iTF.currentTarget, true) --Hide old
					end
					if show then
						statusbar(conditionals.targetChanged.currentTarget.color)
					elseif unitID then
						statusbar(nil, nil, true)
					end
				elseif string.find(k, 'glow') then
					if iTF.currentTarget then
						glows(k,nil, iTF.currentTarget, true) --Hide old
					end
					if show then
						glows(k,conditionals.targetChanged.currentTarget.color)
					elseif unitID then
						glows(k, nil, nil, true)
					end
				end
			end
			iTF.currentTarget = unitID
		end
	end
	if cond == 'maxRange' then
		if conditionals.onUpdate.maxRange then
			local show = false
			--if type(specID.utility) == 'table' then
			--	if IsSpellInRange(specID.utility.min, k) == 0 or IsSpellInRange(specID.utility.min, k) == 0 then
			--		show = true
			--	end
			--elseif IsSpellInRange(specID.utility, unitID) == 0 then
			--	show = true
			--end
			if conditionals.onUpdate.maxRange.invert then
				if type(specID.utility) == 'table' then
					if IsSpellInRange(specID.utility.min, k) == 0 or IsSpellInRange(specID.utility.min, k) == 0 then
						show = true
					end
				elseif IsSpellInRange(specID.utility, unitID) == 0 then
					show = true
				end
			else
				if type(specID.utility) == 'table' then
					if IsSpellInRange(specID.utility.min, k) == 1 or IsSpellInRange(specID.utility.min, k) == 1 then
						show = true
					end
				elseif IsSpellInRange(specID.utility, unitID) == 1 then
					show = true
				end
			end
			for k in pairs(conditionals.onUpdate.maxRange.indicators) do
				if k == 'border' then
					if show then
						border(conditionals.onUpdate.maxRange.color)
					else
						border(nil, nil, true)
					end
				elseif k == 'alpha' then
					if show then
						opacity(conditionals.onUpdate.maxRange.alpha)
					else
						opacity(nil, nil, true)
					end
				elseif k == 'statusbar' then
					if show then
						statusbar(conditionals.onUpdate.maxRange.color)
					else
						statusbar(nil, nil, true)
					end
				elseif string.find(k, 'glow') then
					if show then
						glows(k,conditionals.onUpdate.maxRange.color)
					else
						glows(k, nil, nil, true)
					end
				end
			end
		end
	end
	if cond == 'threat' then
		if conditionals.threat then
			local threat = UnitThreatSituation('player', unitID)
			if threat then
				local threatTable = {
					['gainingAggro'] = 1,
					['losingAggro'] = 2,
					['aggro'] = 3,				
				}
				for threatCond, v in pairs(conditionals.threat) do
					cond = threatCond
					if threatTable[threatCond] == threat then
						local show = false
						if threat and threat > 0 then
							if v.roles == 1 then -- All
								show = true
							elseif v.roles == 2 and specID.tank then --Tank
								show = true
							elseif (not specID.tank) and v.roles == 3 then -- non tanks
								show = true
							end
						end
						for k in pairs(v.indicators) do
							if k == 'border' then
								if show then
									border(v.color)
								else
									border(nil, nil, true)
								end
							elseif k == 'alpha' then
								if show then
									opacity(v.alpha)
								else
									opacity(nil, nil, true)
								end
							elseif k == 'statusbar' then
								if show then
									statusbar(v.color)
								else
									statusbar(nil, nil, true)
								end
							elseif string.find(k, 'glow') then
								if show then
									glows(k,v.color)
								else
									glows(k, nil, nil, true)
								end
							end
						end
					end
				end
			end
		end
	end
	if cond == 'focusTarget' then
		if conditionals.focusUpdate.focusTarget then
			local show = false
			if UnitExists('focus') then
				local nameplate = C_NamePlate.GetNamePlateForUnit('focus')
				if nameplate then
					unitID = nameplate.UnitFrame.unit
					show = true
				end
			end
			for k in pairs(conditionals.focusUpdate.focusTarget.indicators) do
				if k == 'border' then
					if iTF.focusTarget then
						border(nil, iTF.focusTarget, true) --Hide old
					end
					if show then
						border(conditionals.focusUpdate.focusTarget.color)
					elseif unitID then
						border(nil, nil, true)
					end
				elseif k == 'alpha' then
					if iTF.focusTarget then
						opacity(nil,iTF.focusTarget, true) --Hide old
					end
					if show then
						opacity(conditionals.focusUpdate.focusTarget.alpha)
					elseif unitID then
						opacity(nil, nil, true)
					end
				elseif k == 'statusbar' then
					if iTF.focusTarget then
						statusbar(nil,iTF.focusTarget, true) --Hide old
					end
					if show then
						statusbar(conditionals.focusUpdate.focusTarget.color)
					elseif unitID then
						statusbar(nil, nil, true)
					end
				elseif string.find(k, 'glow') then
					if iTF.focusTarget then
						glows(k,nil,iTF.focusTarget, true) --Hide old
					end
					if show then
						glows(k,conditionals.focusUpdate.focusTarget.color)
					elseif unitID then
						glows(k, nil, nil, true)
					end
				end
			end
			iTF.focusTarget = unitID
		end
	end
	if cond and customConditionals[cond] then
		local show = false
		local func, error = getFunction(customConditionals[cond].func, unitID)
		if func(unitID) then
			show = true
		end
		for k in pairs(customConditionals[cond].indicators) do
			if k == 'border' then
				if show then
					border(customConditionals[cond].color)
				else
					border(nil, nil, true)
				end
			elseif k == 'alpha' then
				if show then
					opacity(customConditionals[cond].alpha)
				else
					opacity(nil, nil, true)
				end
			elseif k == 'statusbar' then
				if show then
					statusbar(customConditionals[cond].color)
				else
					statusbar(nil, nil, true)
				end
			elseif string.find(k, 'glow') then
				if show then
					glows(k,customConditionals[cond].color)
				else
					glows(k, nil, nil, true)
				end
			end
		end
	end
end
function iTF:print(...)
	local s = {...}
	local msg = ''
	for _,v in ipairs(s) do
		msg = msg .. v
	end
	if msg == 'help' then
		print('--iTargetingFrames--\n "/itf" to toggle config window.\n "/itf reset" to reset everything to default settings.')
	else
		print('iTF: ' .. msg)
	end
end
function iTF:trimText(unitID)
	while iTF.frames[unitID].text:GetWidth() > iTFConfig.layout.frame.width do
		iTF.frames[unitID].text:SetText(string.sub(iTF.frames[unitID].text:GetText(), 1, -2))
	end
end
function iTF:updateHealth(unitID)
	if iTF.frames[unitID] then
		local hp = UnitHealth(unitID)
		local maxHP = UnitHealthMax(unitID)
		local value = 1
		if hp and maxHP then
			value = hp/maxHP
		end
		iTF.frames[unitID].healthBar:SetValue(value)
		if conditionals.onHealth.custom then
			for k,v in pairs(conditionals.onHealth.custom) do
				if v.func(unitID) then
					updateIndicator('custom' .. k, unitID, v.indicators, true)
				else
					updateIndicator('custom' .. k, unitID, v.indicators)
				end
			end
		end
	end
end
function iTF:updateAuras(unitID)
	local tempAuraTable = {}
	local function sortAuras()
		local id = 1
		for k,v in spairs(tempAuraTable, function(t,a,b) return t[b].expirationTime > t[a].expirationTime end) do
			if id > iTFConfig.layout.icon.max then return end
			if not iTF.frames[unitID].auras[id] then
				iTF:CreateAuraFrame(unitID, id)
			end
			iTF.frames[unitID].auras[id].tex:SetTexture(v.icon)
			iTF.frames[unitID].auras[id].cooldown:SetCooldown(v.start, v.duration)
			iTF.frames[unitID].auras[id]:Show()
			id = id + 1
		end
		for i = id, #iTF.frames[unitID].auras do
			iTF.frames[unitID].auras[i]:Hide()
		end
	end
	for i = 1, 41 do
		local name, _, icon, count, debuffType, duration, expirationTime, _, _, _, spellID = UnitDebuff(unitID, i, 'player')
		if not name then
			sortAuras()
		else
			if not iTFConfig.blacklist[name] and not iTFConfig.blacklist[spellID] then
				table.insert(tempAuraTable, {['icon'] = icon, ['duration'] = duration, ['expirationTime'] = expirationTime, ['count'] = count, ['start'] = expirationTime - duration})
			end
		end
	end
end
function iTF:updateCast(unitID,channel)
	local function GetCast()
		if channel then
			return UnitChannelInfo(unitID)
		else
			return UnitCastingInfo(unitID)
		end
	end
	local name, nameSubtext, text, texture, startTime, endTime, isTradeSkill, castID, notInterruptible = GetCast()
	if name and not iTF.frames[unitID].isMe then
		iTF.frames[unitID].text:SetText(name)
		iTF:trimText(unitID)
		if notInterruptible then
			iTF.frames[unitID].text:SetTextColor(unpack(iTFConfig.layout.colors.text.immune))
		else
			iTF.frames[unitID].text:SetTextColor(unpack(iTFConfig.layout.colors.text.cast))
		end
		iTF.frames[unitID].healthBar:SetHeight(iTFConfig.layout.frame.height-1-iTFConfig.layout.frame.castBarHeight)
		iTF.frames[unitID].castBar:Show()
		--iTF.frames[unitID].castBar:SetMinMaxValues(startTime/1000, endTime/1000)
		local starttime = startTime/1000
		local endtime =  endTime/1000
		iTF.frames[unitID].castBar:SetMinMaxValues(starttime-GetTime(), endtime-GetTime())
		iTF.frames[unitID].castBar:SetScript('OnUpdate', function()
			iTF.frames[unitID].castBar:SetValue(GetTime()-starttime)
		end)
		--iTF.frames[unitID].castBar:SetMinMaxValues(starttime, endtime)
		--iTF.frames[unitID].castBar:SetScript('OnUpdate', function()
		--	iTF.frames[unitID].castBar:SetValue(GetTime())
		--end)
	else
		if UnitExists(unitID) then
			iTF.frames[unitID].text:SetText(UnitName(unitID))
			iTF:trimText(unitID)
			iTF.frames[unitID].text:SetTextColor(unpack(iTFConfig.layout.colors.text.main))
		end
		iTF.frames[unitID].castBar:SetScript('OnUpdate', nil)
		iTF.frames[unitID].castBar:SetValue(0)
		iTF.frames[unitID].castBar:Hide()
		iTF.frames[unitID].healthBar:SetHeight(iTFConfig.layout.frame.height)
	end
end
function iTF:CreateAuraFrame(unitID, id)
	--Create frame
	iTF.frames[unitID].auras[id] = CreateFrame('frame',nil,iTF.frames[unitID])
	iTF.frames[unitID].auras[id]:SetSize(iTFConfig.layout.icon.width,iTFConfig.layout.icon.height)
	--Backdrop
	iTF.frames[unitID].auras[id]:SetBackdrop(iTF.backdrop)
	iTF.frames[unitID].auras[id]:SetBackdropColor(unpack(iTFConfig.layout.colors.backdrop.bg))
	iTF.frames[unitID].auras[id]:SetBackdropBorderColor(unpack(iTFConfig.layout.colors.backdrop.border))
	iTF.frames[unitID].auras[id]:SetFrameStrata('MEDIUM')
	if id == 1 then
		iTF.frames[unitID].auras[id]:SetPoint('BOTTOMLEFT', iTF.frames[unitID], 'BOTTOMLEFT', 1, 1)
	else
		iTF.frames[unitID].auras[id]:SetPoint('LEFT', iTF.frames[unitID].auras[id-1], 'RIGHT', 2, 0)
	end
	--Icon
	iTF.frames[unitID].auras[id].tex = iTF.frames[unitID].auras[id]:CreateTexture()
	iTF.frames[unitID].auras[id].tex:SetAllPoints(iTF.frames[unitID].auras[id])
	--Cooldown
	iTF.frames[unitID].auras[id].cooldown = CreateFrame('Cooldown', nil, iTF.frames[unitID].auras[id], 'CooldownFrameTemplate')
	iTF.frames[unitID].auras[id].cooldown:SetAllPoints(iTF.frames[unitID].auras[id])
	iTF.frames[unitID].auras[id].cooldown:SetReverse(true)
	iTF.frames[unitID].auras[id].cooldown:SetHideCountdownNumbers(false)
end
function iTF:getUFPos(id)
	local r = id%iTFConfig.layout.colS
	local x = math.floor(id/iTFConfig.layout.colS)*(iTFConfig.layout.frame.width+iTFConfig.layout.frame.spacing)
	local y = (r-1)*(iTFConfig.layout.frame.height+iTFConfig.layout.frame.spacing)
	
	if r == 0 then
		x = (math.floor(id/iTFConfig.layout.colS)-1)*(iTFConfig.layout.frame.width+iTFConfig.layout.frame.spacing)
		y = (iTFConfig.layout.colS-1)*(iTFConfig.layout.frame.height+iTFConfig.layout.frame.spacing)
	end
	if iTFConfig.layout.grow == 'TOPRIGHT' then
		return -x, -y
	elseif iTFConfig.layout.grow == 'TOPLEFT' then
		return x, -y
	elseif iTFConfig.layout.grow ==  'BOTTOMRIGHT' then
		return -x, y
	else --BOTTOMLEFT
		return x, y
	end
	return posX, posY
end
function iTF:hideAll(frame, loop)
	local function hide(k)
		iTF.frames[k]:SetAlpha(1)
		iTF.frames[k].alphaInUse = false
		iTF.frames[k].topFrame.glowLeft:Hide()
		iTF.frames[k].topFrame.glowLeft.inUse = false
		iTF.frames[k].topFrame.glowRight:Hide()
		iTF.frames[k].topFrame.glowRight.inUse = false
		iTF.frames[k].topFrame.glowTop:Hide()
		iTF.frames[k].topFrame.glowTop.inUse = false
		iTF.frames[k].topFrame.glowBottom:Hide()
		iTF.frames[k].topFrame.glowBottom.inUse = false
		iTF.frames[k]:SetBackdropBorderColor(unpack(iTFConfig.layout.colors.statusbar.border))
		iTF.frames[k].borderInUse = false
		iTF.frames[k].healthBar:SetStatusBarColor(unpack(iTFConfig.layout.colors.statusbar.main))
		iTF.frames[k].healthBar.inUse = false
	end
	if loop then
		for k in pairs(iTF.frames) do
			hide(k)
		end
	else
		hide(frame)
	end
end
function iTF:updateRaidIcon(unitID)
	local function checkUnit(unitID)
		local raidIcon = GetRaidTargetIndex(unitID)
		if raidIcon and raidIcon < 9 then
			iTF.frames[unitID].raidIcon:SetTexture('Interface\\TargetingFrame\\UI-RaidTargetingIcon_'..raidIcon)
			iTF.frames[unitID].raidIcon:Show()
		else
			iTF.frames[unitID].raidIcon:Hide()
		end
	end
	if unitID then
		checkUnit(unitID)
	else
		for k in pairs(iTF.frames) do
			if UnitExists(k) then
				checkUnit(k)
			end
		end
	end
end
function iTF:updateUnitID(unitID)
	iTF:hideAll(unitID)
	iTF.frames[unitID].guid = UnitGUID(unitID)
	iTF.frames[unitID].waitingFor = {
		['border'] = {},
		['alpha'] = {},
		['statusbar'] = {},
		['glowLeft'] = {},
		['glowRight'] = {},
		['glowTop'] = {},
		['glowBottom'] = {},
	}
	if UnitIsUnit(unitID, 'player') then
		iTF.frames[unitID].isMe = true
	else
		iTF.frames[unitID].isMe = false
	end
	if iTFConfig.layout.colors.classColor and UnitIsPlayer(unitID) then
		iTF.frames[unitID].isPlayer  = {}
		local _, class = UnitClass(unitID)
		iTF.frames[unitID].isPlayer.color = RAID_CLASS_COLORS[class]
		iTF.frames[unitID].isPlayer.class = class
		iTF.frames[unitID].healthBar:SetStatusBarColor(iTF.frames[unitID].isPlayer.color.r, iTF.frames[unitID].isPlayer.color.g, iTF.frames[unitID].isPlayer.color.b, iTFConfig.layout.colors.statusbar.main.a)
	else
		iTF.frames[unitID].isPlayer = nil
		iTF.frames[unitID].healthBar:SetStatusBarColor(unpack(iTFConfig.layout.colors.statusbar.main))
	end
	iTF:updateHealth(unitID)
	iTF:updateAuras(unitID)
	iTF:updateRaidIcon(unitID)
	iTF.frames[unitID].text:SetText(UnitName(unitID))
	iTF:trimText(unitID)
	iTF:updateCast(unitID)
	for _,v in pairs(conditionals) do --force update all
		for cond, _ in pairs(v) do
			updateIndicator(unitID,cond)
		end
	end
end
function iTF:CreateNew(unitID, i)
	if not iTF.frames then
		iTF.frames = {}
	end
	local posX, posY = iTF:getUFPos(i)
	iTF.frames[unitID] = CreateFrame('BUTTON', 'iTF_'..unitID, iTF.mainFrame, 'SecureUnitButtonTemplate,SecureHandlerEnterLeaveTemplate,SecureHandlerShowHideTemplate')
	iTF.frames[unitID]:SetFrameStrata('LOW')
	iTF.frames[unitID]:SetAttribute('unit', unitID)
	iTF.frames[unitID]:SetAttribute('_onenter',[[local script = self:GetAttribute('itf_onenter'); if script then self:Run(script);end]])
	iTF.frames[unitID]:SetAttribute('_onleave',[[self:ClearBindings()]])
	iTF.frames[unitID]:SetAttribute('_onshow',[[if self:IsUnderMouse() then local script = self:GetAttribute('itf_onenter');if script then self:Run(script);end;end]])
	iTF.frames[unitID]:SetAttribute('_onhide',[[if self:IsUnderMouse() then self:ClearBindings();end]])
	iTF.frames[unitID]:SetSize(iTFConfig.layout.frame.width+2,iTFConfig.layout.frame.height+2)
	iTF.frames[unitID]:SetPoint(iTFConfig.layout.grow, iTF.mainFrame, iTFConfig.layout.grow, posX, posY)
	
	iTF.frames[unitID].nameplateID = i
	
	--Backdrop
	iTF.frames[unitID]:SetBackdrop(iTF.backdrop)
	iTF.frames[unitID]:SetBackdropColor(unpack(iTFConfig.layout.colors.statusbar.backdrop))
	iTF.frames[unitID]:SetBackdropBorderColor(unpack(iTFConfig.layout.colors.statusbar.border))
	
	--Health Bar
	iTF.frames[unitID].healthBar = CreateFrame('StatusBar', nil,iTF.frames[unitID])
	iTF.frames[unitID].healthBar:SetStatusBarTexture(iTFConfig.layout.statusbar.texture)
	iTF.frames[unitID].healthBar:SetWidth(iTFConfig.layout.frame.width)
	iTF.frames[unitID].healthBar:SetHeight(iTFConfig.layout.frame.height)
	iTF.frames[unitID].healthBar:SetPoint('BOTTOMLEFT', iTF.frames[unitID], 'BOTTOMLEFT', 1,1)
	iTF.frames[unitID].healthBar:SetStatusBarColor(unpack(iTFConfig.layout.colors.statusbar.main))
	iTF.frames[unitID].healthBar:SetMinMaxValues(0,1)
	
	--Name text
	iTF.frames[unitID].text = iTF.frames[unitID].healthBar:CreateFontString()
	iTF.frames[unitID].text:SetFont(iTFConfig.layout.text.font, iTFConfig.layout.text.size, 'OUTLINE')
	iTF.frames[unitID].text:SetPoint(iTFConfig.layout.text.pos, iTF.frames[unitID], iTFConfig.layout.text.pos, iTFConfig.layout.text.x,iTFConfig.layout.text.y)
	iTF.frames[unitID].text:SetText('test')
	iTF.frames[unitID].text:SetTextColor(unpack(iTFConfig.layout.colors.text.main))
	
	--Cast Bar
	iTF.frames[unitID].castBar = CreateFrame('StatusBar', nil,iTF.frames[unitID])
	iTF.frames[unitID].castBar:SetStatusBarTexture(iTFConfig.layout.castBar.texture)
	iTF.frames[unitID].castBar:SetWidth(iTFConfig.layout.frame.width)
	iTF.frames[unitID].castBar:SetHeight(iTFConfig.layout.frame.castBarHeight)
	iTF.frames[unitID].castBar:SetPoint('TOPLEFT', iTF.frames[unitID], 'TOPLEFT', 1,-1)
	iTF.frames[unitID].castBar:SetStatusBarColor(unpack(iTFConfig.layout.colors.statusbar.cast))
	iTF.frames[unitID].castBar:SetMinMaxValues(0,1)
	iTF.frames[unitID].castBar:SetValue(0)
	iTF.frames[unitID].castBar:Hide()
	
	--Glow
	iTF.frames[unitID].topFrame = CreateFrame('frame', nil, iTF.frames[unitID]) -- anchor frame
	iTF.frames[unitID].topFrame:SetAllPoints(iTF.frames[unitID])
	iTF.frames[unitID].topFrame:SetFrameStrata('MEDIUM')
	--Top
	iTF.frames[unitID].topFrame.glowTop = iTF.frames[unitID].topFrame:CreateTexture()
	iTF.frames[unitID].topFrame.glowTop:SetPoint('TOP', iTF.frames[unitID].topFrame, 'TOP', 0,-1)
	iTF.frames[unitID].topFrame.glowTop:SetSize(iTFConfig.layout.frame.width, 8)
	iTF.frames[unitID].topFrame.glowTop:SetTexture('Interface\\AddOns\\iTargetingFrames\\media\\hGlow')
	iTF.frames[unitID].topFrame.glowTop:SetTexCoord(0, 60/64, 0, 1)
	iTF.frames[unitID].topFrame.glowTop:SetVertexColor(1,1,1,0.8)
	iTF.frames[unitID].topFrame.glowTop:Hide()
	--Bottom
	iTF.frames[unitID].topFrame.glowBottom = iTF.frames[unitID].topFrame:CreateTexture()
	iTF.frames[unitID].topFrame.glowBottom:SetPoint('BOTTOM', iTF.frames[unitID].topFrame, 'BOTTOM', 0,1)
	iTF.frames[unitID].topFrame.glowBottom:SetSize(iTFConfig.layout.frame.width, 8)
	iTF.frames[unitID].topFrame.glowBottom:SetTexture('Interface\\AddOns\\iTargetingFrames\\media\\hGlow')
	iTF.frames[unitID].topFrame.glowBottom:SetTexCoord(0, 60/64, 1, 0)
	iTF.frames[unitID].topFrame.glowBottom:SetVertexColor(1,0,1,0.8)
	iTF.frames[unitID].topFrame.glowBottom:Hide()
	--Left
	iTF.frames[unitID].topFrame.glowLeft = iTF.frames[unitID].topFrame:CreateTexture()
	iTF.frames[unitID].topFrame.glowLeft:SetPoint('LEFT', iTF.frames[unitID].topFrame, 'LEFT', 1,0)
	iTF.frames[unitID].topFrame.glowLeft:SetSize(8, iTFConfig.layout.frame.height)
	iTF.frames[unitID].topFrame.glowLeft:SetTexture('Interface\\AddOns\\iTargetingFrames\\media\\hGlow')
	iTF.frames[unitID].topFrame.glowLeft:SetTexCoord(60/64,0,0,0,60/64,1,0,1)
	iTF.frames[unitID].topFrame.glowLeft:SetVertexColor(1,1,0,0.8)
	iTF.frames[unitID].topFrame.glowLeft:Hide()
	--Right
	iTF.frames[unitID].topFrame.glowRight = iTF.frames[unitID].topFrame:CreateTexture()
	iTF.frames[unitID].topFrame.glowRight:SetPoint('RIGHT', iTF.frames[unitID].topFrame, 'RIGHT', -1,0)
	iTF.frames[unitID].topFrame.glowRight:SetSize(8, iTFConfig.layout.frame.height)
	iTF.frames[unitID].topFrame.glowRight:SetTexture('Interface\\AddOns\\iTargetingFrames\\media\\hGlow')
	iTF.frames[unitID].topFrame.glowRight:SetTexCoord(0,1,60/64,1,0,0,60/64,0)
	iTF.frames[unitID].topFrame.glowRight:SetVertexColor(0.5,1,0,0.8)
	iTF.frames[unitID].topFrame.glowRight:Hide()
	--[[
	--Top left
	iTF.frames[unitID].topFrame.glowTopLeft = iTF.frames[unitID].topFrame:CreateTexture()
	iTF.frames[unitID].topFrame.glowTopLeft:SetPoint('TOPLEFT', iTF.frames[unitID].topFrame, 'TOPLEFT', 1,-1)
	iTF.frames[unitID].topFrame.glowTopLeft:SetSize(24, 24)
	iTF.frames[unitID].topFrame.glowTopLeft:SetTexture('Interface\\AddOns\\iTargetingFrames\\media\\cGlow')
	iTF.frames[unitID].topFrame.glowTopLeft:SetTexCoord(1, 0, 0, 1)
	iTF.frames[unitID].topFrame.glowTopLeft:SetVertexColor(0,1,1,1)
	--Top right
	iTF.frames[unitID].topFrame.glowTopRight = iTF.frames[unitID].topFrame:CreateTexture()
	iTF.frames[unitID].topFrame.glowTopRight:SetPoint('TOPRIGHT', iTF.frames[unitID].topFrame, 'TOPRIGHT', -1,-1)
	iTF.frames[unitID].topFrame.glowTopRight:SetSize(24, 24)
	iTF.frames[unitID].topFrame.glowTopRight:SetTexture('Interface\\AddOns\\iTargetingFrames\\media\\cGlow')
	iTF.frames[unitID].topFrame.glowTopRight:SetTexCoord(0, 1, 0, 1)
	iTF.frames[unitID].topFrame.glowTopRight:SetVertexColor(1,0,1,1)
	--Bottom left
	iTF.frames[unitID].topFrame.glowBottomLeft = iTF.frames[unitID].topFrame:CreateTexture()
	iTF.frames[unitID].topFrame.glowBottomLeft:SetPoint('BOTTOMLEFT', iTF.frames[unitID].topFrame, 'BOTTOMLEFT', 1, 1)
	iTF.frames[unitID].topFrame.glowBottomLeft:SetSize(8, 8)
	iTF.frames[unitID].topFrame.glowBottomLeft:SetTexture('Interface\\AddOns\\iTargetingFrames\\media\\cGlow')
	iTF.frames[unitID].topFrame.glowBottomLeft:SetTexCoord(1, 0, 1, 0)
	iTF.frames[unitID].topFrame.glowBottomLeft:SetVertexColor(1,0,1,1)
	--Bottom right
	iTF.frames[unitID].topFrame.glowBottomRight = iTF.frames[unitID].topFrame:CreateTexture()
	iTF.frames[unitID].topFrame.glowBottomRight:SetPoint('BOTTOMRIGHT', iTF.frames[unitID].topFrame, 'BOTTOMRIGHT', -1, 1)
	iTF.frames[unitID].topFrame.glowBottomRight:SetSize(8, 8)
	iTF.frames[unitID].topFrame.glowBottomRight:SetTexture('Interface\\AddOns\\iTargetingFrames\\media\\cGlow')
	iTF.frames[unitID].topFrame.glowBottomRight:SetTexCoord(0, 1, 1, 0)
	iTF.frames[unitID].topFrame.glowBottomRight:SetVertexColor(1,1,0,1)
	--]]
	--Raid Icon
	iTF.frames[unitID].raidIcon = iTF.frames[unitID].topFrame:CreateTexture()
	iTF.frames[unitID].raidIcon:SetSize(iTFConfig.layout.raidIcon.size, iTFConfig.layout.raidIcon.size)
	iTF.frames[unitID].raidIcon:SetPoint(iTFConfig.layout.raidIcon.pos, iTF.frames[unitID].topFrame, iTFConfig.layout.raidIcon.pos, iTFConfig.layout.raidIcon.x,iTFConfig.layout.raidIcon.y)
	iTF.frames[unitID].raidIcon:Hide()
	--Auras
	iTF.frames[unitID].auras = {}
	--Hide & Scripts
	iTF.frames[unitID]:Hide()
	iTF.frames[unitID]:HookScript('OnShow',function()
		iTF:updateUnitID(unitID)
	end)
	iTF.frames[unitID]:HookScript('OnHide', function()
		iTF.frames[unitID].isPlayer = nil
	end)
	RegisterUnitWatch(iTF.frames[unitID])
end
local onUpdateTotal = 0
function iTF:OnUpdate(elapsed)
	onUpdateTotal = onUpdateTotal + elapsed
	if onUpdateTotal >= 0.2 then
		for k,v in pairs(iTF.frames) do
			if UnitExists(k) then
				if v.guid ~= UnitGUID(k) then
					--iTF:updateUnitID(k)
				--elseif not iTF.frames[k].isMe then
				else
					for cond,_ in pairs(conditionals.onUpdate) do
						updateIndicator(k, cond)
					end
				--else
					--iTF:hideAll(k)
				--	iTF.frames[k]:SetAlpha(1)
				end
			--elseif iTF.frames[k].isMe then
				--iTF.frames[k]:SetAlpha(1)
			end
		end
		onUpdateTotal = 0
	end
end
function iTF:updateFrames(toUpdate)
	if not toUpdate or toUpdate == 'size' then
		--Update background size
		local width = (iTFConfig.layout.frame.width+iTFConfig.layout.frame.spacing)*math.ceil(30/iTFConfig.layout.colS)
		local height = (iTFConfig.layout.frame.height+iTFConfig.layout.frame.spacing)*iTFConfig.layout.colS
		iTF.mainFrame:SetSize(width, height)
		local i = 1
		for k,v in pairs(iTF.frames) do
			local posX, posY = iTF:getUFPos(v.nameplateID)
			if UnitExists(k) then
				iTF.frames[k].text:SetText(UnitName(k))
			end
			iTF:trimText(k)
			iTF.frames[k]:SetSize(iTFConfig.layout.frame.width+2,iTFConfig.layout.frame.height+2)
			iTF.frames[k]:ClearAllPoints()
			iTF.frames[k]:SetPoint(iTFConfig.layout.grow, iTF.mainFrame, iTFConfig.layout.grow, posX, posY)
			iTF.frames[k].healthBar:SetWidth(iTFConfig.layout.frame.width)
			iTF.frames[k].healthBar:SetHeight(iTFConfig.layout.frame.height)
			iTF.frames[k].castBar:SetWidth(iTFConfig.layout.frame.width)
			iTF.frames[k].castBar:SetHeight(iTFConfig.layout.frame.castBarHeight)
			iTF.frames[k].topFrame.glowTop:SetWidth(iTFConfig.layout.frame.width)
			iTF.frames[k].topFrame.glowBottom:SetWidth(iTFConfig.layout.frame.width)
			iTF.frames[k].topFrame.glowLeft:SetHeight(iTFConfig.layout.frame.height)
			iTF.frames[k].topFrame.glowRight:SetHeight(iTFConfig.layout.frame.height)
			i = i + 1
		end
	end
	if not toUpdate or toUpdate == 'pos' then
		iTF.mainFrame:ClearAllPoints()
		iTF.mainFrame:SetPoint(iTFConfig.layout.grow, UIParent, 'BOTTOMLEFT', iTFConfig.layout.anchor.x, iTFConfig.layout.anchor.y)
		local width = (iTFConfig.layout.frame.width+iTFConfig.layout.frame.spacing)*math.ceil(40/iTFConfig.layout.colS)
		local height = (iTFConfig.layout.frame.height+iTFConfig.layout.frame.spacing)*iTFConfig.layout.colS
		iTF.mainFrame:SetSize(width, height)
		for k,v in pairs(iTF.frames) do
			local posX, posY = iTF:getUFPos(v.nameplateID)
			iTF.frames[k]:ClearAllPoints()
			iTF.frames[k]:SetPoint(iTFConfig.layout.grow, iTF.mainFrame, iTFConfig.layout.grow, posX, posY)
		end
	end
	if not toUpdate or toUpdate == 'glow' then
		for k in pairs(iTF.frames) do
			iTF.frames[k].topFrame.glowTop:Hide()
			iTF.frames[k].topFrame.glowBottom:Hide()
			iTF.frames[k].topFrame.glowLeft:Hide()
			iTF.frames[k].topFrame.glowRight:Hide()
		end
	end
	if not toUpdate or toUpdate == 'text' then
		for k,v in pairs(iTF.frames) do
			iTF.frames[k].text:SetFont(iTFConfig.layout.text.font, iTFConfig.layout.text.size, 'OUTLINE')
			if UnitExists(k) then
				iTF.frames[k].text:SetText(UnitName(k))
				iTF:trimText(k)
			end
			iTF.frames[k].text:ClearAllPoints()
			iTF.frames[k].text:SetPoint(iTFConfig.layout.text.pos, iTF.frames[k], iTFConfig.layout.text.pos, iTFConfig.layout.text.x, iTFConfig.layout.text.y)
			iTF:updateCast(k)
		end
	end
	if not toUpdate or toUpdate == 'statusBarColor' then
		for k in pairs(iTF.frames) do
			if iTFConfig.layout.colors.classColor and UnitExists(k) and UnitIsPlayer(k) then
				local _, class = UnitClass(k)
				local color = RAID_CLASS_COLORS[class]
				iTF.frames[k].healthBar:SetStatusBarColor(color.r, color.g, color.b, iTFConfig.layout.colors.statusbar.main.a)
			else
				iTF.frames[k].healthBar:SetStatusBarColor(unpack(iTFConfig.layout.colors.statusbar.main))
			end
		end
	end
	if not toUpdate or toUpdate == 'statusbar' then
		for k in pairs(iTF.frames) do
			iTF.frames[k].healthBar:SetStatusBarTexture(iTFConfig.layout.statusbar.texture)
			iTF.frames[k].castBar:SetStatusBarTexture(iTFConfig.layout.castBar.texture)
		end
	end
	if not toUpdate or toUpdate == 'frameColor' then
		for k in pairs(iTF.frames) do
			iTF.frames[k]:SetBackdropColor(unpack(iTFConfig.layout.colors.statusbar.backdrop))
			iTF.frames[k]:SetBackdropBorderColor(unpack(iTFConfig.layout.colors.statusbar.border))
		end
	end
	if not toUpdate or toUpdate == 'castBarColor' then
		for k in pairs(iTF.frames) do
			iTF.frames[k].castBar:SetStatusBarColor(unpack(iTFConfig.layout.colors.statusbar.cast))
		end
	end
	if not toUpdate or toUpdate == 'target' then
		updateIndicator('currentTarget')
	end
	if not toUpdate or toUpdate == 'raidIcon' then
		for k in pairs(iTF.frames) do
			iTF.frames[k].raidIcon:SetSize(iTFConfig.layout.raidIcon.size,iTFConfig.layout.raidIcon.size)
			iTF.frames[k].raidIcon:ClearAllPoints()
			iTF.frames[k].raidIcon:SetPoint(iTFConfig.layout.raidIcon.pos, iTF.frames[k].topFrame, iTFConfig.layout.raidIcon.pos, iTFConfig.layout.raidIcon.x, iTFConfig.layout.raidIcon.y)
		end
	end
	if not toUpdate or toUpdate == 'auraIcon' then
		local to = {['LEFT'] = 'Right', ['RIGHT'] = 'Left', ['TOP'] = 'Down', ['BOTTOM'] = 'Up'}
		local x,y = 0,0
		local grow = iTFConfig.layout.icon.grow
		if grow == 'LEFT' then
			x,y = 2,0
		elseif grow == 'RIGHT' then
			x,y = -2,0
		elseif grow == 'TOP' then
			x,y = 0,-2
		else
			x,y = 0,2
		end
		
		for k in pairs(iTF.frames) do
			for id = 1, #iTF.frames[k].auras do
				iTF.frames[k].auras[id]:SetSize(iTFConfig.layout.icon.width, iTFConfig.layout.icon.height)
				iTF.frames[k].auras[id]:ClearAllPoints()
				if id == 1 then
					iTF.frames[k].auras[id]:SetPoint(iTFConfig.layout.icon.pos, iTF.frames[k], iTFConfig.layout.icon.pos, iTFConfig.layout.icon.x, iTFConfig.layout.icon.y)
				else
					iTF.frames[k].auras[id]:SetPoint(grow, iTF.frames[k].auras[id-1], to[grow], x, y)
				end
			end
		end
	end
	if not toUpdate or toUpdate == 'conditionals' then
		conditionals = nil
		conditionals = {
			['onUpdate'] = {},
			['onAura'] = {},
			['onCast'] = {},
			['onHealth'] = {},
			['targetChanged'] = {},
			['threat'] = {},
			['focusUpdate'] = {},
		}
		customCondtionals = nil
		customCondtionals = {}
		iTF:hideAll(nil,true)
		for k,v in pairs(iTFConfig.layout.conditionals) do
			if v.enable then
				indicatorPrios[k] = v.weight
				if v.update then --nil check for earlier versions TO DO: remove on v0.5 or something
					conditionals[v.update][k] = v
				else
					iTF:print('Update function not found for ',k,', adding ',iTF.indicatorUpdateFuncs[k],'.')
					iTFConfig.layout.conditionals[k].update = iTF.indicatorUpdateFuncs[k]
					conditionals[iTF.indicatorUpdateFuncs[k]][k] = v
				end
			end
		end
		for k,v in pairs(iTFConfig.custom) do
			if v.enable then
				indicatorPrios[k] = v.weight
				if v.update then --nil check for earlier versions TO DO: remove on v0.5 or something
					conditionals[v.update][k] = v
					customConditionals[k] = v
				else
					iTF:print('Update function not found for ',k,', adding ',iTF.indicatorUpdateFuncs[k],'.')
					v.update = iTF.indicatorUpdateFuncs[k]
				end
			end
		end
		for unitID in pairs(iTF.frames) do
			if UnitExists(unitID) then
				iTF:updateUnitID(unitID)
			end
		end
	end
	if not toUpdate or toUpdate == 'bindings' then
		for k in pairs(iTF.frames) do
			local keys = {}
			local mouseKeys = {'LeftButtonDown'}
			local mousewheel = false
			local function applyBindings(t)
				if t then
					local function isMouse(button)
						if button == 'LeftButton' then
							return 1
						elseif button == 'RightButton' then
							return 2
						elseif button == 'MiddleButton' then
							return 3
						elseif button == 'BUTTON4' then
							return 4
						elseif button == 'BUTTON5' then
							return 5
						else
							return false
						end
					end
					for key,v in pairs(t) do
						local unModifiedKey = key
						local modifier = ''
						local mouseKey = isMouse(unModifiedKey)
						if mouseKey then
							--alt-ctrl-shift-type
							if key:find('ALT') then
								modifier = modifier .. 'alt-'
								unModifiedKey = unModifiedKey:gsub('ALT%-', '')
							end
							if key:find('CTRL') then
								modifier = modifier .. 'ctrl-'
								unModifiedKey = unModifiedKey:gsub('CTRL%-', '')
							end
							if key:find('SHIFT') then
								modifier = modifier .. 'shift-'
								unModifiedKey = unModifiedKey:gsub('SHIFT%-', '')
							end
							table.insert(mouseKeys, unModifiedKey .. 'Down')
						end
						if not mousewheel then
							if key:find('MouseWheel') then
								mousewheel = true
							end
						end
						if not v.text then --Target, Focus
							if mouseKey then
								iTF.frames[k]:SetAttribute(modifier..'type'..mouseKey,v.type)
							else
								iTF.frames[k]:SetAttribute(string.format([[type-itf%s]],key),v.type)
							end
						else
							if mouseKey then
								iTF.frames[k]:SetAttribute(modifier..'type'..mouseKey,v.type)
							else
								iTF.frames[k]:SetAttribute(string.format([[type-itf%s]],key),v.type)
							end
							if v.type == 'macro' then
								if mouseKey then
									iTF.frames[k]:SetAttribute(modifier..'macrotext'..mouseKey,v.text)
								else
									iTF.frames[k]:SetAttribute(string.format([[macrotext-itf%s]],key),v.text)
								end
							else
								if mouseKey then
									iTF.frames[k]:SetAttribute(string.format([[%s%s%s]],modifier,v.type,mouseKey),v.text)
								else
									iTF.frames[k]:SetAttribute(string.format([[%s-itf%s]],v.type,key),v.text)
								end
							end							
						end
						if not mouseKey then
							keys[key] = true
						end
					end
				end
			end
			--prio: general < class < spec
			applyBindings(iTFConfig.bindings.general)
			applyBindings(iTFConfig.bindings[iTF.class].b)
			applyBindings(iTFConfig.bindings[iTF.class][specID.specID])
			local bind = ''
			for key in pairs(keys) do
				bind = bind .. string.format([[self:SetBindingClick(true, '%s', self, 'itf%s');]], key,key)
			end
			iTF.frames[k]:RegisterForClicks(unpack(mouseKeys))
			if mousewheel then
				iTF.frames[k]:EnableMouseWheel(true)
			else
				iTF.frames[k]:EnableMouseWheel(false)
			end
			iTF.frames[k]:SetAttribute('itf_onenter', bind)
		end
	end
end
function iTF:CreateMainFrame()
	iTF.mainFrame = CreateFrame('frame', nil, UIParent)
	local width = (iTFConfig.layout.frame.width+iTFConfig.layout.frame.spacing)*math.ceil(40/iTFConfig.layout.colS)
	local height = (iTFConfig.layout.frame.height+iTFConfig.layout.frame.spacing)*iTFConfig.layout.colS
	iTF.mainFrame:SetSize(width, height)
	iTF.mainFrame:SetPoint(iTFConfig.layout.grow, UIParent, 'BOTTOMLEFT', iTFConfig.layout.anchor.x, iTFConfig.layout.anchor.y)
	iTF.mainFrame.tex = iTF.mainFrame:CreateTexture()
	iTF.mainFrame.tex:SetAllPoints(iTF.mainFrame)
	iTF.mainFrame.tex:SetColorTexture(0.5,0.5,0.5,0.5)
	iTF.mainFrame.tex:Hide()
	for i = 1, 40 do --Create 30 frames
		iTF:CreateNew('nameplate' .. i,i)
	end
	iTF.mainFrame:SetScript('OnUpdate', iTF.OnUpdate)
end
function addon:ADDON_LOADED(addonName)
	if addonName == 'iTargetingFrames' then
		iTFConfig = iTFConfig or {}

		iTF.class = select(3,UnitClass('player')) --number
		conditionals = nil
		conditionals = {
			['onUpdate'] = {},
			['onAura'] = {},
			['onCast'] = {},
			['onHealth'] = {},
			['targetChanged'] = {},
		}
		iTF:LoadDefaults()
		addon:UnregisterEvent('ADDON_LOADED')
		iTF:CreateMainFrame()
		iTF:updateFrames('conditionals')
		local events = {'UNIT_HEALTH','UNIT_HEALTH_FREQUENT','UNIT_SPELLCAST_START','UNIT_SPELLCAST_STOP','UNIT_SPELLCAST_INTERRUPTED','UNIT_SPELLCAST_CHANNEL_START','UNIT_SPELLCAST_CHANNEL_STOP','UNIT_AURA','ENCOUNTER_START','ENCOUNTER_STOP','RAID_TARGET_UPDATE','PLAYER_LOGIN','PLAYER_SPECIALIZATION_CHANGED','PLAYER_TARGET_CHANGED','UNIT_THREAT_LIST_UPDATE','PLAYER_FOCUS_CHANGED','PLAYER_REGEN_DISABLED', 'NAME_PLATE_UNIT_ADDED', 'NAME_PLATE_UNIT_REMOVED'}
		for i = 1, #events do
			addon:RegisterEvent(events[i])
		end
		--[[
		addon:RegisterEvent('UNIT_HEALTH')
		addon:RegisterEvent('UNIT_HEALTH_FREQUENT')
		addon:RegisterEvent('UNIT_SPELLCAST_START')
		addon:RegisterEvent('UNIT_SPELLCAST_STOP')
		addon:RegisterEvent('UNIT_SPELLCAST_INTERRUPTED')
		addon:RegisterEvent('UNIT_SPELLCAST_CHANNEL_START')
		addon:RegisterEvent('UNIT_SPELLCAST_CHANNEL_STOP')
		addon:RegisterEvent('UNIT_AURA')
		addon:RegisterEvent('ENCOUNTER_START')
		addon:RegisterEvent('ENCOUNTER_STOP')
		addon:RegisterEvent('RAID_TARGET_UPDATE')
		addon:RegisterEvent('PLAYER_LOGIN')
		addon:RegisterEvent('PLAYER_SPECIALIZATION_CHANGED')
		addon:RegisterEvent('PLAYER_TARGET_CHANGED')
		addon:RegisterEvent('UNIT_THREAT_LIST_UPDATE')
		addon:RegisterEvent('PLAYER_FOCUS_CHANGED')
		--]]
	end
end
function addon:UNIT_HEALTH_FREQUENT(unitID)
	if iTF.frames[unitID] then
		for k,_ in pairs(conditionals.onHealth) do
			updateIndicator(unitID, k)
		end
		iTF:updateHealth(unitID)
	end
end
function addon:UNIT_HEALTH(unitID)
	if iTF.frames[unitID] then
		for k,_ in pairs(conditionals.onHealth) do
			updateIndicator(unitID, k)
		end
		iTF:updateHealth(unitID)
	end
end
function addon:UNIT_AURA(unitID)
	if iTF.frames[unitID] then
		for k,_ in pairs(conditionals.onAura) do
			updateIndicator(unitID, k)
		end
		iTF:updateAuras(unitID)
	end
end
function addon:UNIT_SPELLCAST_START(unitID)
	if iTF.frames[unitID] then
		for k,_ in pairs(conditionals.onCast) do
			updateIndicator(unitID, k)
		end
		iTF:updateCast(unitID)
	end
end
function addon:UNIT_SPELLCAST_STOP(unitID)
	if iTF.frames[unitID] then
		for k,_ in pairs(conditionals.onCast) do
			updateIndicator(unitID, k)
		end
		iTF:updateCast(unitID)
	end
end
function addon:UNIT_SPELLCAST_INTERRUPTED(unitID)
	if iTF.frames[unitID] then
		for k,_ in pairs(conditionals.onCast) do
			updateIndicator(unitID, k)
		end
		iTF:updateCast(unitID)
	end
end
function addon:UNIT_SPELLCAST_CHANNEL_START(unitID)
	if iTF.frames[unitID] then
		for k,_ in pairs(conditionals.onCast) do
			updateIndicator(unitID, k)
		end
		iTF:updateCast(unitID, true)
	end
end
function addon:UNIT_SPELLCAST_CHANNEL_STOP(unitID)
	if iTF.frames[unitID] then
		for k,_ in pairs(conditionals.onCast) do
			updateIndicator(unitID, k)
		end
		iTF:updateCast(unitID, true)
	end
end
function addon:ENCOUNTER_START()
	iTF.encounterInProgress = true
end
function addon:ENCOUNTER_STOP()
	iTF.encounterInProgress = false
end
function addon:RAID_TARGET_UPDATE()
	iTF:updateRaidIcon()
end
function addon:PLAYER_LOGIN()
	iTF.class = select(3,UnitClass('player')) --number
	iTF.specID = GetSpecializationInfo(GetSpecialization())
	specID = {
		['specID'] = iTF.specID,
		['utility'] = iTF.spells.range[iTF.specID].utility,
		['interrupt'] = iTF.spells.range[iTF.specID].interrupt,
		['dps'] = iTF.spells.range[iTF.specID].dps,
	}
	if not iTFConfig.bindings[iTF.class] then
		iTFConfig.bindings[iTF.class] = {['b'] = {}}
	end
	if not iTFConfig.bindings[iTF.class][specID.specID] then
		iTFConfig.bindings[iTF.class][specID.specID] = {}
	end
	iTF:updateFrames('bindings')
end
function addon:PLAYER_SPECIALIZATION_CHANGED()
	iTF.specID = GetSpecializationInfo(GetSpecialization())
	if iTF.specID then
		specID = {
			['specID'] = iTF.specID,
			['utility'] = iTF.spells.range[iTF.specID].utility,
			['interrupt'] = iTF.spells.range[iTF.specID].interrupt,
			['dps'] = iTF.spells.range[iTF.specID].dps,
		}
		if not iTFConfig.bindings[iTF.class] then
		iTFConfig.bindings[iTF.class] = {['b'] = {}}
		end
		if not iTFConfig.bindings[iTF.class][specID.specID] then
			iTFConfig.bindings[iTF.class][specID.specID] = {}
		end
		iTF:updateFrames('bindings')
	end
end
function addon:UNIT_THREAT_LIST_UPDATE(unitID)
	if iTF.frames[unitID] then
		for k,_ in pairs(conditionals.threat) do
			updateIndicator(unitID, k)
		end
		--updateIndicator(unitID, 'threat')
	end
end
function addon:PLAYER_TARGET_CHANGED()
	updateIndicator(nil, 'currentTarget')
	for k,_ in pairs(conditionals.targetChanged) do
		updateIndicator(nil, k)
	end
end
function addon:PLAYER_FOCUS_CHANGED()
	for k,_ in pairs(conditionals.focusUpdate) do
			updateIndicator(nil, k)
	end
	--updateIndicator(nil, 'focusTarget')
end
function addon:PLAYER_REGEN_DISABLED()
	iTF:toggleConfig(true)
end
function addon:NAME_PLATE_UNIT_ADDED(unitID)
	if iTF.frames[unitID] then
		iTF:updateUnitID(unitID)
	end
end
function addon:NAME_PLATE_UNIT_REMOVED(unitID)
	if iTF.frames[unitID] then
		iTF.frames[unitID].isPlayer = nil
	end
end

SLASH_ITF1 = "/itf"
SLASH_ITF2 = '/itargetingframes'
SlashCmdList["ITF"] = function(msg)
	msg = string.lower(msg:gsub('^%s*(.-)%s*$', '%1'))
	if string.len(msg) == 0 then
		msg = nil
	end
	if msg and msg == 'reset' then
		iTF:LoadDefaults(true)
	elseif msg then
		iTF:print('help')
	else
		if not InCombatLockdown() then
			iTF:toggleConfig()
		else
			iTF:print(L.combatLockdown)
		end
	end
end